#include "Approx.h"


//===========================================================================================
//===========================================================================================

//////////////////
// CONSTRUCTORS //
//////////////////

//===========================================================================================

//Building of an object approx thanks to a matrice file fm

approx::approx(short order,double length,char* fm,char* fa,ofstream& fr,const gsl_rng * r)
{  
  int k;
  double rho;
  int i,j,l,m,error;
  double q,t;
  int tMU;

  k=order;
  t=length;
   
  //Printing datas
  if(VERBOSE)
    {
      cout<<"\x1b[001;034m###########\x1b[000m"<<endl
	  <<"\x1b[001;034m## Datas ##\x1b[000m"<<endl
	  <<"\x1b[001;034m###########\x1b[000m"<<endl<<endl;
      
      cout <<"Sequence length = "<<t<<endl
	   <<"Order of the Markov model = "<<k<<endl
	   <<"Alphabet = "<<fa<<endl<<endl;
    }
  fr<<"###########"<<endl
    <<"## Datas ##"<<endl
    <<"###########"<<endl<<endl;
 
  fr<<"Sequence length = "<<t<<endl
    <<"Order of the Markov model = "<<k<<endl
    <<"Alphabet = "<<fa<<endl<<endl;
  
  //INITIALIZE PRIVATE MEMBERS
  _order=k; 
  _alphabet = new Translator(fa); 
  _sa=_alphabet->tell_alphabet_size();
  tMU=int(pow(double(_sa),double(k)));  //tMU is not an approx member!
  _mu=new double[_sa];
  _MU=new double[tMU];
  _sPI=tMU*_sa;
  _PI=new double[_sPI]; 
  _K=0;
  _nu=0;
  _length=t;
 
  //_mu just used in probawll to approximate probability of a letter by (1/ta) 
  //when the length of the word is smaller than the order of the Markov model
  for(l=0;l<_sa;l++)
    {
      _mu[l]=double(1)/double(_sa);
    }

  //Read a transition matrix from file fm
  //Verify sum on each line is equal to 1
  error=0; 
  ifstream tr(fm);
  char line[256];
  tr.getline(line,256);
  tr.getline(line,256);
  tr.getline(line,256);
  tr.getline(line,256);
  for(i=0; i<tMU; i++)
    {
      q=0;
      for(j=0;j<_sa;j++)
 	{
 	  tr>>_PI[_sa*i+j];
 	  q=q+_PI[_sa*i+j];	     
 	}
      if (fabs(q-1)>0.000000001)
 	{
 	  error=error+1;	     
 	}	 
    }
  tr.close(); 

  //Building an object Markov thanks to the matrix file
  Markov mk(fm);
  mk.compute_stat_law(true);

  for(i=0; i<_sPI; i++)
    {
      _PI[i]=mk.markov_matrix()[i];
    }
  //Computing of the stationary probability
  if(VERBOSE)
    { 
      cout<<"\x1b[001;034m"
	  <<"####################################################\x1b[000m"<<endl<<"\x1b[001;034m"
	  <<"## Transition matrix and stationary probabilities ##\x1b[000m"<<endl<<"\x1b[001;034m"
	  <<"####################################################\x1b[000m"<<endl<<endl;
    }
  fr<<"####################################################"<<endl
    <<"## Transition matrix and stationary probabilities ##"<<endl
    <<"####################################################"<<endl
    <<endl;
  if(VERBOSE)
    { 
      cout<<"\x1b[036mTransition matrix : \x1b[000m"<<endl;
    }
  fr<<"Transition matrix : "<<endl;
  for(i=0;i<tMU;i++)
    { 
      for(j=0;j<_sa;j++)
	{
	  if(VERBOSE)
	    { 
	      cout<<_PI[_sa*i+j]<<" ";
	    }
	  fr<<_PI[_sa*i+j]<<" ";
	}
      if(VERBOSE)
	{ 
	  cout<<endl;
	}
      fr<<endl;
    }
  //Stationary probabilities 
  if(VERBOSE)
    { 
      cout<<endl<<"\x1b[036mStationary probabilities : \x1b[000m"<<endl;
    }
  fr<<endl<<"Stationary probabilities : "<<endl;
  for(i=0;i<tMU;i++) 
    {
      //_MU[i]= mk.Mu(i);
      _MU[i]= mk.stat_law()[i];
      if(VERBOSE)
	{ 
	  cout<<_MU[i]<<" ";
	}
      fr<<_MU[i]<<" ";
      if (((i+1)%_sa)==0)
	{
	  if(VERBOSE)
	    { 
	      cout<<endl;
	    }
	  fr<<endl;
	}
    }
 
  //Computing of the second eigenvalue of M
  //after having "transformed" _Pi in a square matrix!
  double **M = new double*[tMU];
  for(i=0;i<tMU;i++)
    {
      M[i]= new double[tMU];
      for(j=0;j<tMU;j++)
	{
	  M[i][j]=0;
	}
    }
  for (i=0; i<tMU; i++)
    {
      int ind = i % (tMU/_sa) ;
      ind *= _sa ;
      for (j=0; j<_sa;j++)
	{
	  M[i][ind+j]=_PI[i*_sa + j] ;	   
	}
    }
  double *eigenr= new double[tMU];
  double *eigeni = new double[tMU] ;
  gsl_vector *eigenmod = gsl_vector_alloc(tMU) ;
  int nerr ;
  //Diagonalization
  GetEigen(tMU,M,eigenr,eigeni,&nerr) ;
  //Module of eigenvalues
  for(i=0; i<tMU;i++) 
    gsl_vector_set(eigenmod,i,hypot(eigenr[i],eigeni[i]));
  gsl_sort_vector(eigenmod) ;
  
  //Eigenvalues of Pi
  if(VERBOSE)
    { 
      cout<<endl<<"\x1b[036mEigenvalues of Pi :\x1b[000m"<<endl;
    }
  for(i=0; i<tMU; i++) 
    {
      if(VERBOSE)
	{ 
	  cout<<gsl_vector_get(eigenmod,i)<<endl;
	}
    }
  
  rho = gsl_vector_get(eigenmod,tMU-2);
      
  delete[] eigenr;
  delete[] eigeni;
  gsl_vector_free(eigenmod);
  delete[] M;

  
  /*======================*/
  /* K=1/mumin and nu=rho */
  /*======================*/
  
  _nu = rho;
  _K = double(1)/(inf(_MU,tMU));
  fr<<endl<<"nu = Second eigenvalue of PI : "<<endl<<_nu<<endl;  
  fr<<endl<<"Value of K : "<<endl<<_K<<endl; 
  
}
//===========================================================================================

//Building of an object approx thanks to a sequence file fm

approx::approx(short order, ofstream& fr,char* fs,char* fa,const gsl_rng * r)
{ 
  int k;
  double t;
  double rho=0;
  int i,j,l,m;
  int tMU;
  
  k=order;
  SequenceSet seqs(fs,fa,order);
  t=double(seqs.tell_length());

  //Printing datas
  if(VERBOSE)
    {
      cout<<"\x1b[001;034m###########\x1b[000m"<<endl
	  <<"\x1b[001;034m## Datas ##\x1b[000m"<<endl
	  <<"\x1b[001;034m###########\x1b[000m"<<endl<<endl;
      
      cout <<"Sequence length = "<<t<<endl
	   <<"Order of the Markov model = "<<k<<endl
	   <<"Alphabet = "<<fa<<endl<<endl;
    }
  fr<<"###########"<<endl
    <<"## Datas ##"<<endl
    <<"###########"<<endl<<endl;
 
  fr<<"Sequence length = "<<t<<endl
    <<"Order of the Markov model = "<<k<<endl
    <<"Alphabet = "<<fa<<endl<<endl;
  
  //INITIALIZE PRIVATE MEMBERS
  _order=k;
  _alphabet = new Translator(fa); 
  _sa=_alphabet->tell_alphabet_size();
  tMU=int(pow(double(_sa),double(k)));  //tMU is not an approx member!
  _mu=new double[_sa];
  _MU=new double[tMU];
  _sPI=tMU*_sa;
  _PI=new double[_sPI];
  _K=0;
  _nu=0;
  _length=t;

  //_mu just used in probawll to approximate probability of a letter by (1/ta) 
  //when the length of the word is smaller than the order of the Markov model
  for(l=0;l<_sa;l++)
    {
      _mu[l]=double(1)/double(_sa);
    }
  
  //Computing of the stationary probability
  if(VERBOSE)
    { 
      cout<<"\x1b[001;034m"
	  <<"####################################################\x1b[000m"<<endl<<"\x1b[001;034m"
	  <<"## Transition matrix and stationary probabilities ##\x1b[000m"<<endl<<"\x1b[001;034m"
	  <<"####################################################\x1b[000m"<<endl<<endl;
    }
  fr<<"####################################################"<<endl
    <<"## Transition matrix and stationary probabilities ##"<<endl
    <<"####################################################"<<endl
    <<endl;
  
  //Building an object Markov thanks to the sequence file
  Markov mk(seqs);
  mk.compute_stat_law(true);
   
  //Transition matrix is estimated in building of mk
  for(i=0; i<_sPI; i++)
    {
      _PI[i]=mk.Pi(i);
    }
  if(VERBOSE)
    { 
      cout<<"\x1b[036mTransition matrix : \x1b[000m"<<endl;
    }
  fr<<"Transition matrix : "<<endl;
  for(i=0;i<tMU;i++)
    { 
      for(j=0;j<_sa;j++)
	{
	  fr<<_PI[_sa*i+j]<<" ";
	  if(VERBOSE)
	    { 
	      cout<<_PI[_sa*i+j]<<" ";
	    }
	}
      fr<<endl;
      if(VERBOSE)
	{ 
	  cout<<endl;
	}
    }
  //Stationary probabilities 
  if(VERBOSE)
    { 
      cout<<endl<<"\x1b[036mStationary probabilities  : \x1b[000m"<<endl; 
    }
  fr<<endl<<"Stationary probabilities  : "<<endl; 
  for(i=0;i<tMU;i++) 
    {
     
      //_MU[i]= mk.Mu(i);
      _MU[i]= mk.stat_law()[i];
      fr<<_MU[i]<<" "; 
      if(VERBOSE)
	{ 
	  cout<<_MU[i]<<" ";
	}
      if(((i+1)%_sa)==0)
	{
	  fr<<endl;
	  if(VERBOSE)
	    { 
	      cout<<endl;
	    }
	}
    }

  //Computing of the second eigenvalue of M
  //after having "transformed" _Pi in a square matrix!
  double **M = new double*[tMU];
  for(i=0;i<tMU;i++)
    {
      M[i]= new double[tMU];
      for(j=0;j<tMU;j++)
	{
	  M[i][j]=0;
	}
    }
  for (i=0; i<tMU; i++)
    {
      int ind = i % (tMU/_sa) ;
      ind *= _sa ;
      for (j=0; j<_sa;j++)
	{
	  M[i][ind+j]=_PI[i*_sa + j] ;	   
	}
    }
  double *eigenr= new double[tMU];
  double *eigeni = new double[tMU] ;
  gsl_vector *eigenmod = gsl_vector_alloc(tMU) ;
  int nerr ;
  //Diagonalization
  GetEigen(tMU,M,eigenr,eigeni,&nerr) ;
  //Module of eigenvalues
  for(i=0; i<tMU;i++) 
    gsl_vector_set(eigenmod,i,hypot(eigenr[i],eigeni[i]));
  gsl_sort_vector(eigenmod) ;
  
  //Eigenvalues of Pi
  if(VERBOSE)
    { 
      cout<<endl<<"\x1b[036mEigenvalues of Pi :\x1b[000m"<<endl;
    }
  for(i=0; i<tMU; i++) 
    {
      if(VERBOSE)
	{ 
	  cout<<gsl_vector_get(eigenmod,i)<<endl;
	}
    }
  
  rho = gsl_vector_get(eigenmod,tMU-2);
      
  delete[] eigenr;
  delete[] eigeni;
  gsl_vector_free(eigenmod); 
  delete[] M;

  
  /*======================*/
  /* K=1/mumin and nu=rho */
  /*======================*/

  _nu = rho;
  _K = double(1)/(inf(_MU,tMU));
  fr<<endl<<"nu = Second eigenvalue of PI : "<<endl<<_nu<<endl;  
  fr<<endl<<"Value de K : "<<endl<<_K<<endl; 
}


//===========================================================================================
//===========================================================================================


////////////////
// DESTRUCTOR //
////////////////

//===========================================================================================

approx::~approx()
{
  delete _alphabet;
  delete[] _PI;
  delete[] _mu; 
  delete[] _MU;
}


//===========================================================================================
//===========================================================================================

/////////////
// METHODS //
/////////////

//===========================================================================================

//Printing transition matrix

void approx::printPI()
{
  int i,j,power; 
  power=int(pow(double(_sa),double(_order)));
  for(i=0;i<power;i++)
    { 
      for(j=0;j<_sa;j++)
	{ 
	  cout<<_PI[_sa*i+j]<<" ";
	}
      cout<<endl;
    }
}

//===========================================================================================

//Printing transition matrix in file fm

void approx::printPI(ofstream& fm)
{
  int i,j,power; 
  power=int(pow(double(_sa),double(_order)));
  for(i=0;i<power;i++)
    { 
      for(j=0;j<_sa;j++)
	{ 
	  fm<<_PI[_sa*i+j]<<" ";
	}
      fm<<endl;
    }
}

//===========================================================================================

//Return the sequence length

double approx::tell_length ()
{
  return _length;
}

//===========================================================================================


//Probability of the w last letters of A of size n thanks to _PI and _mu

double approx::probawll(int w, char* A, int n)
{ 
  
  int i,j,c;
  double p1,p2,p;
  char* B=new char[_order+1+1];
  char* D=new char[_order+1];
  B[_order+1]='\0';
  D[_order]='\0';
  p1=1;
  p2=1;
  int len=0;
  if(w>=_order)
    {
      for (c=n-w;c<n-w+_order;c++)
	{
	  D[c-n+w]=A[c];
	} 
      //Stationary probabilities
      p1=_MU[_alphabet->word_to_coded_int(D,len)]; 
         
      for(j=n-w;j<n-_order;j++)
	{
	  for (c=j;c<=j+_order;c++)
	    {
	      B[c-j]=A[c];
	    } 
	  
	  //B is a word of size order+1 and we want calculate 
	  //the probability of the last letter of B knowing the order first letters.
	  //Index i such that _PI[i]=proba(A[a]|A[d]A[d+1]...A[a-1]) 
	  //is  _alphabet->word_to_coded_int(B,len)
	  p2=p2*_PI[_alphabet->word_to_coded_int(B,len)];
	}    
    }
  else
    {
      //If w<order, then probability of the w last letters is approximated by (1/ta)^k
      for(i=0;i<w;i++)
	{
	  //word_to_coded_int have a char * argument 
	  char * Une=new char[1+1];
	  Une[0]=A[n-w+i];
	  Une[1]='\0';
	  p1=p1*_mu[_alphabet->word_to_coded_int(Une,len)];  
	  delete[] Une;
	}
    }
  p=p1*p2;
  delete[] B;
  delete[] D; 
  return p;
}

//===========================================================================================

//Function phi 

double approx::phi(double l)
{ 
  double r;
  r = _K*pow(_nu,l);
  return r;
}

//===========================================================================================

//epsilon 1
//A word A of size n 

double approx::epsilon_one(char* A,int n)    
{
  double inf1,inf2,epsilon1;
  int i;
  int pp, nA, rA;
  periods(A,n,pp,nA,rA);

  inf1=double(n+rA)*probawll(1,A,n)+phi(double(nA)-double(1));
  for(i=2;i<=nA;i++)
    {
      inf2=double(n+rA)*probawll(i,A,n)+phi(double(nA)-double(i));     
      if (inf2<inf1)
	{
	  inf1=inf2;
	}
    }
  epsilon1=inf1;
  return epsilon1;
}


//===========================================================================================

//See report

double approx::compute_la(double p, int n)
{
  double x,y,la;
  x=-p*p;
  y=log(_nu);
  la=log(x/(_K*y))/y;
  //Function increase from la, then inf is obtained for n 
  //la must be greater than n 
  if(la<n)
    {
      la=n;
    } 
  //Function decrease before la, then inf is obtained for 1/p 
  //la must be smaller than 1/p
  if(la>1/p)
    {
      la=1/p;
    }
  return la;
}

//===========================================================================================

//See report

double approx::epsilon_two(double p, int n)
{
  double eps;
  double la = compute_la(p,n);
  eps=la*p + phi(la)/p + phi(n);
  return eps;
}

//===========================================================================================

//inf(Dp(A)+phi(D))

double approx::varepsilon(double p, int n)
{
  //Computing Delta
  double x,y,D;
  x=-p;
  y=log(_nu);
  D=log(x/(_K*y))/y;
  //Function increase from la, then inf is obtained for n 
  //la must be greater than n 
  if(D<n)
    {
      D=n;
    } 
  //Function decrease before la, then inf is obtained for 1/p 
  //la must be smaller than 1/p
  if(D>1/p)
    {
      D=1/p;
    }
  
  //Computing varepsilon
  double eps;
  eps=D*p+phi(D);
  return eps;
 
}

//===========================================================================================


//Return proposition phi (error in 0)

double approx::proposition_phi(double p, int n, double t, char* A)
{
  double Ca,Cb,X1,X2,Cx1;
  Ca=24;
  Cb=25;
  X1=1/(2+n+_K+(_K/(1-_nu))); 
  if(X1<1-11*epsilon_one(A,n))
    {
      X1=1-11*epsilon_one(A,n);
    }
  X2=2;
  Cx1=11;  

  double t1,t2,t3,t4,T;
  t1=(8+2*Cb+Ca+t*p*Ca)*varepsilon(p,n)*exp(-X1*t*p);
  t2=t*p*Cx1*epsilon_one(A,n)*exp(-X1*t*p);
  T=t1+t2;
  return T;
}

//===========================================================================================

//epsilon 1
//A word A of size n 

double approx::psiepsilon_one(char* A,int n)    
{ 
  double inf1,inf2,epsilon1;
  int i;
  int pp, nA, rA;
  periods(A,n,pp,nA,rA);

  inf1=double(n+rA)*probawll(1,A,n)*(1+phi(double(nA)-double(1)));
  for(i=2;i<=nA;i++)
    {
      inf2=double(n+rA)*probawll(i,A,n)*(1+phi(double(nA)-double(i)));     
      if (inf2<inf1)
	{
	  inf1=inf2;
	}
    }
  epsilon1=inf1;
  return epsilon1;
  
}


//===========================================================================================

//Return proposition psi (error in 0)
//Min{Article, Thesis}

double approx::proposition_psi(double p, int n, double t, char* A)
{
  double Ca,Cb,X1,X2,Cx1;
  Ca=24;
  Cb=25;
  Cx1=11;
  X1=1/(2+n+_K+(_K/(1-_nu)));
  if(X1<1-Cx1*psiepsilon_one(A,n))
    {
      X1=1-Cx1*psiepsilon_one(A,n);
    }
  X2=2;
  
  double t1,t2,t3,t4,T1,T2,T;
  t1=(8+2*Cb+Ca+t*p*Ca)*varepsilon(p,n)*exp(-X1*t*p);
  t2=t*p*Cx1*psiepsilon_one(A,n)*exp(-X1*t*p);
  T1=t1+t2;
 
  //Thesis :
  double C= 1-phi(double(3*n));
  double lnu=log(_nu);
  double x= -(5+2*phi(n))*p*C;
  double y=4*_K*lnu;
  double z=log(x/y)/lnu;
  double d=deltacs(p,n);
  double eps= (5+2*phi(n))*d*p+4*phi(d)/(1-phi(3*n));
  double CS=t*p*(psiepsilon_one(A,n)+eps);
  
  if(CS<T1)
    {T=CS;}
  else
    {T=T1;}
  return T;
}

//===========================================================================================
   
//Return alphabet

Translator * approx::give_alpha() const
{
  Translator * alpha(_alphabet);
  return alpha;
}

//===========================================================================================

//Give the jth word (length lj) of file wf
//A contains this word

int approx::give_word (char * A, char* wf,int j)
{
  int nbm,i;
  char T[256];
 
  ifstream mo;
  mo.open(wf);  
  mo>>nbm;
  //Go to the line j+1 (line of the jth word)
  for(i=0; i<j+1; i++)
    {
      mo.getline(T,256);
    }
  //jth word in A
  mo>>A;
  mo.close();

  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                  PHI MIXING                 @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@     Overrepresented words      @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::error_k
(int k, double s1, double lx, double le, double la, double p1, int n, double p, double par)
{ 
  int g,z;
  double A,B,C,S,pbk,t1,t2,t3,t4,lk; 
  double t=_length;  
  double res;

  if (k==0)
    {
      res=p1;
    }
  else
    {		  
      //P(Bk)
      A=0;
      B=-log(double(k));
      C=0;
      S=0;
      pbk=0;			  
      for(z=1;z<k;z++)
	{
	  A=A+log(t-z+1);
	  B=B+log(double(k-z+1));
	  C=C+log(double(z));
	  S=A+B-2*C+z*log(p+phi(la))+(k-z)*le;
	  S=exp(S);
	  pbk=pbk+S;
	}
      //Other terms
      lk=k*log(par)-s1;
      lk=exp(lk);
      t1=log(5.0)+le+k*lx-s1;
      t1=exp(t1);
      t2=lk*(k+1)*p1/k;
      t3=lk*p*(2*la+2*n)*(k+1)*exp(p*((k+1)*(2*la+2*n)-t))/k;
      t4 = lk*abs(pow((t-k-4*n)/t,k)-1)*exp(-par)/k;
      //RESULT		  
      res=pbk+t1+t2+t3+t4; 
    }
  return res;
}

//===========================================================================================

int approx::print_error_k(char* wf,int k,ofstream& fr)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m##########################\x1b[000m"<<endl
	  <<"\x1b[032m##### PHI ERROR IN K #####\x1b[000m"<<endl
	  <<"\x1b[032m##########################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"##########################"<<endl
    <<"##### PHI ERROR IN K #####"<<endl
    <<"##########################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPhi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Phi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p1,e1,e2,e,le,la,b,res,x,lx,par;
  double s1;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of the words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //Word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
  	}
      else
	{
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1
	      e1=epsilon_one(V,n[i]);	      	     	      
	      //proposition phi
	      p1=proposition_phi(p,n[i],t,V);
	      if (k==0)
		{
		  res=p1;
		  if(VERBOSE)
		    {
		      cout<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
		    }
		  fr<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
		}
	      else
		{
		  //epsilon2
		  e2=epsilon_two(p,n[i]);
		  //Error proposition 21
		  e=e1+e2;		 
		  if(e>=1)
		    {
		      if(VERBOSE)
			{
			  cout<<"Epsilon too large"<<endl<<endl;
			}
		      fr<<"Epsilon too large"<<endl<<endl;
		    }
		  else
		    {
		      le=log(e);	      
		      //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		      la=compute_la(p,n[i]);
		      x=(p+phi(la))*t;		  
		      lx=log(x);	
		      b=double(2)*t/double(n[i]);   
		      //Too large number of occurrences :
		      if (k>b)
			{
			  if(VERBOSE)
			    {
			      cout<<"Impossible : "<<endl
				  <<"a word of size "<<n[i]
				  << " and of period greater than "<< n[i]/2<<endl
				  <<"can not appear "<<k<<" times in a sequence of length "
				  <<t<<endl;
			      cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			    }
			  fr<<"Impossible : "<<endl
				  <<"a word of size "<<n[i]
				  << " and of period greater than "<< n[i]/2<<endl
				  <<"can not appear "<<k<<" times in a sequence of length "
				  <<t<<endl;
			  fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      else
			{
			  s1=0;
			  for (g=1;g<k;g++)
			    {
			      s1=s1+log(double(g));
			    }			  
			  res = error_k(k,s1,lx,le,la,p1,n[i],p,par);
			  if(VERBOSE)
			    {
			      cout<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
			    }
			  fr<<"RESULTAT (error in "<<k<<") = "<<res<<endl<<endl;
			}
		    }
		}
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}

//===========================================================================================

double approx::sum_errors_k
(int k,double s1,double lx, double le, double la, double p1, int n, double p, double par)
{
  int j;
  double t=_length;
  double b= double(2)*t/double(n);
  double pr=pow(double(10),double(-300));
  double res=0;
  double result=pow(double(10),double(300));
  for (j=k;(j<=b) && (result>pr);j++)
    {
      result = error_k(j,s1,lx,le,la,p1,n,p,par);
      res = res+result;
      s1=s1+log(double(j));
    }
  res = res + (b-j)*pr;
  return res;
}

//===========================================================================================

int approx::print_sum_errors_k(char* wf,int k,ofstream& fr)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m#################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PHI SUM ERRORS OVER K #####\x1b[000m"<<endl
	  <<"\x1b[032m#################################\x1b[000m"<<endl<<endl;
    }
  fr<<endl;
  fr<<"#################################"<<endl
    <<"##### PHI SUM ERRORS OVER K #####"<<endl
    <<"#################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPhi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Phi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p1,e1,e2,e,le,la,b,res,x,lx,par,s1;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of the words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //Word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
  	}
      else
	{
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1,
	      e1=epsilon_one(V,n[i]);	      	     	      
	      //proposition phi
	      p1=proposition_phi(p,n[i],t,V);
	      //epsilon2.
	      e2=epsilon_two(p,n[i]);
	      //Error proposition 21
	      e=e1+e2;		 
	      if(e>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		}
	      else
		{
		  le=log(e);	      
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  la=compute_la(p,n[i]);
		  x=(p+phi(la))*t;		  
		  lx=log(x);	
		  b=double(2)*t/double(n[i]);   
		  //Too large number of occurrences :
		  if (k>b)
		    {
		      if(VERBOSE)
			{
			  cout<<"Impossible : "<<endl
			      <<"a word of size "<<n[i]
			      << " and of period greater than "<< n[i]/2<<endl
			      <<"can not appear "<<k<<" times in a sequence of length "
			      <<t<<endl;
			  cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      fr<<"Impossible : "<<endl
			<<"a word of size "<<n[i]
			<< " and of period greater than "<< n[i]/2<<endl
			<<"can not appear "<<k<<" times in a sequence of length "
			<<t<<endl;
		      fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
		    }
		  else
		    {
		      s1=0;
		      for (g=1;g<k;g++)
			{
			  s1=s1+log(double(g));
			}	
		      res = sum_errors_k(k,s1,lx,le,la,p1,n[i],p,par);
		      double sm =sum_app(k,b,par);
		      if(VERBOSE)
			{
			  cout<<"RESULT (sum errors) = "<<res<<endl
			      <<"P("<<k<<" <= X <= "<<b<<"),"
			      <<" if X follows Poisson ("<<par<<") = "
			      <<sm<<endl<<endl;
			}
		      fr<<"RESULT (sumerrors) = "<<res<<endl
			<<"P("<<k<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
			<<sm<<endl<<endl;
		    }
		}	       
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}

//===========================================================================================

int approx::print_u(char* wf, double ds,ofstream& fr, ofstream& fu)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m#########################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PHI U (overrepresented words) #####\x1b[000m"<<endl
	  <<"\x1b[032m#########################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"#########################################"<<endl
    <<"##### PHI U (overrepresented words) #####"<<endl
    <<"#########################################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPhi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Phi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"####################################################"<<endl
    <<"##### u for phi mixing (overrepresented words) #####"<<endl
    <<"####################################################"<<endl<<endl;

  double t = _length;
 
  int k=0;
  int i,j,g,z;
  double p,p1,e1,e2,e,le,la,b,res,result,x,s1,s2;
  double par,lx,pr,sm;
  pr=pow(double(10),double(-300));

  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {
      //word i+1
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      if(inBn2(V,n[i]))
  	{ 
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible"<<endl;
  	}
      else
	{
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible"<<endl;
	    }  	  
	  else
	    {
	      //epsilon1
	      e1=epsilon_one(V,n[i]); 
	      //epsilon2	      
	      e2=epsilon_two(p,n[i]);
	      //proposition phi
	      p1=proposition_phi(p,n[i],t,V);
	      //Error proposition 21
	      e=e1+e2;
	      if(e>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		  fu<<"impossible"<<endl;
		}
	      else
		{
		  le=log(e);
		  la=compute_la(p,n[i]);
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=(p+phi(la))*t;	     	     
		  lx=log(x);
		  b=double(2)*t/double(n[i]);     		  
		  res= pow(double(10),double(300));
		  //Search u from parameter (the mean)
		  k=int(par)+1; 
		  s1=0;
		  for (g=1;g<k;g++)
		    {
		      s1=s1+log(double(g));
		    }
		  //TO RETURN U
		  result=pow(double(10),double(300));
		  while (res>ds)
 		    {	
		      j=k;
 		      res=sum_app(j,b,par);
 		      sm=res;
		      s2=s1;
		      if(sm<ds)
			{
			  while ((j<=b) && (result>pr) && (res<ds))
			    {		      
			      result=error_k(j,s2,lx,le,la,p1,n[i],p,par);
			      res=res+result; 
			      s2=s2+log(double(j)); 
			      j++;
			    }
			  if (result<pr)
			    {
			      res=res+(b-j+1)*pr;			 
			    }
			}
 		      s1=s1+log(double(k));  		      
 		      k++;		       
 		    }
		  if(VERBOSE)
		    {
		      cout<<"RESULT (sum errors) = "<<res<<endl
			  <<"P("<<k-1<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
			  <<sm<<endl<<endl;
		    }
		  fr<<"RESULT (sumerrors) = "<<res<<endl
		    <<"P("<<k-1<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
		    <<sm<<endl<<endl;
		  fu<<k-1<<endl;
		}
	    }  			       
	}
      delete[] V;    
    }
  delete[] n; 
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                 PHI MIXING                  @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@     underrepresented words     @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::k_sum_errors
(int k,double lx, double le, double la, double p1, int n, double p, double par)
{  
  int j;
  double res=0;
  double result;
  double s1=0;
  result = error_k(0,s1,lx,le,la,p1,n,p,par);
  res = res+result;
  for (j=1;j<k+1;j++)
    {
      result = error_k(j,s1,lx,le,la,p1,n,p,par);
      res = res+result;
      s1=s1+log(double(j));
    }
  return res;
}

//===========================================================================================
 
int approx::print_k_sum_errors(char* wf,int k,ofstream& fr)
{  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m#################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PHI SUM ERRORS UNDER K#####\x1b[000m"<<endl
	  <<"\x1b[032m#################################\x1b[000m"<<endl<<endl;
    }
  fr<<endl;
  fr<<"##################################"<<endl
    <<"##### PHI SUM ERRORS UNDER K #####"<<endl
    <<"##################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPhi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Phi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p1,e1,e2,e,le,la,b,res,x,lx,par;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	}
      else
	{
	  //sa proba,
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1
	      e1=epsilon_one(V,n[i]);	      	     	      
	      //proposition phi
	      p1=proposition_phi(p,n[i],t,V);
	      //epsilon2
	      e2=epsilon_two(p,n[i]);
	      //Error proposition 21
	      e=e1+e2;		 
	      if(e>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		}
	      else
		{
		  le=log(e);	      
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  la=compute_la(p,n[i]);
		  x=(p+phi(la))*t;		  
		  lx=log(x);	
		  b=double(2)*t/double(n[i]);   
		  //Too large number of occurrences :
		  if (k>b)
		    {
		      if(VERBOSE)
			{
			  cout<<"Impossible : "<<endl
			      <<"a word of size "<<n[i]
			      << " and of period greater than "<< n[i]/2<<endl
			      <<"can not appear "<<k<<" times in a sequence of length "
			      <<t<<endl;
			  cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      fr<<"Impossible : "<<endl
			<<"a word of size "<<n[i]
			<< " and of period greater than "<< n[i]/2<<endl
			<<"can not appear "<<k<<" times in a sequence of length "
			<<t<<endl;
		      fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
		    }
		  else
		    {		      	
		      res = k_sum_errors(k,lx,le,la,p1,n[i],p,par);
		      double sm =sum_app(0,k,par);
		      if(VERBOSE)
			{
			  cout<<"RESULT (sum errors) = "<<res<<endl
			      <<"P(0 <= X <= "<<k<<"),"
			      <<" if X follows Poisson ("<<par<<") = "
			      <<sm<<endl<<endl;
			}
		      fr<<"RESULT (sum errors) = "<<res<<endl
			<<"P(0 <= X <= "<<k<<"), if X follows Poisson ("<<par<<") = "
			<<sm<<endl<<endl;
		    }
		}	       
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}
  
//===========================================================================================

int approx::u_print(char* wf, double ds,ofstream& fr, ofstream& fu)
{  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m##########################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PHI U (underrepresented words) #####\x1b[000m"<<endl
	  <<"\x1b[032m##########################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"##########################################"<<endl
    <<"##### PHI U (underrepresented words) #####"<<endl
    <<"##########################################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPhi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Phi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"#####################################################"<<endl
    <<"##### u for phi mixing (underrepresented words) #####"<<endl
    <<"#####################################################"<<endl<<endl;

  double t = _length;
 
  int k=0;
  int i,j,g,z;
  double p,p1,e1,e2,e,le,la,res,result,x,s1;
  double par,lx,sm;
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words  
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      if(inBn2(V,n[i]))
  	{
 	 if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible"<<endl;
  	}
      else
	{
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible"<<endl;
	    }  	  
	  else
	    {
	      //epsilon1
	      e1=epsilon_one(V,n[i]); 
	      //epsilon2	      
	      e2=epsilon_two(p,n[i]);
	      //proposition phi
	      p1=proposition_phi(p,n[i],t,V);
	      //Error proposition 21
	      e=e1+e2;
	      if(e>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		  fu<<"impossible"<<endl;
		}
	      else
		{
		  le=log(e);
		  la=compute_la(p,n[i]);
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=(p+phi(la))*t;	     	     
		  lx=log(x);   		  
		  res=0;
		  //Search u from 0
		  k=0; 
		  s1=0;
		  result=error_k(0,s1,lx,le,la,p1,n[i],p,par);
		  res=res+result+sum_app(0,0,par);
		  k++;
		  //TO RETURN U
		  while (res<ds)
 		    {
 		      res=sum_app(0,k,par);
 		      sm=res;
		      if(sm<ds)
			{
			  j=1;
			  while ((j<=k) && (res<ds))
			    {		      
			      result=error_k(j,s1,lx,le,la,p1,n[i],p,par);
			      res=res+result; 
			      s1=s1+log(double(j)); 
			      j++;
			    }			
			}		      		      
 		      k++;		       
 		    }
		  if(VERBOSE)
		    {
		      cout<<"RESULT (sum errors) = "<<res<<endl
			  <<"P(0 <= X <= "<<k-1<<"), if X follows Poisson ("<<par<<") = "
			  <<sm<<endl<<endl;
		    }
		  fr<<"RESULTAT (sum errors) = "<<res<<endl
		    <<"P(0 <= X <= "<<k-1<<"), if X follows Poisson ("<<par<<") = "
		    <<sm<<endl<<endl;
		  fu<<k-1<<endl;
		}
	    }  			       
	}
      delete[] V;    
    }
  delete[] n; 
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                  PSI MIXING                 @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@      Overrepresented word      @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::psi_error_k
(int k, double s1, double lx, double le, double p3, int n, double p, double par)
{ 
  int g,z;
  double A,B,C,S,pbk,t1,t2,t3,t4,lk; 
  double t=_length;  
  double res;

  if (k==0)
    {
      res=p3;
    }
  else
    {		  
      //P(Bk)
      A=0;
      B=-log(double(k));
      C=0;
      S=0;
      pbk=0;			  
      for(z=1;z<k;z++)
	{
	  A=A+log(t-z+1);
	  B=B+log(double(k-z+1));
	  C=C+log(double(z));
	  S=A+B-2*C+z*log(p*(1+phi(n)))+(k-z)*le;
	  S=exp(S);
	  pbk=pbk+S;
	}
      pbk=pbk*exp(-(t-(3*k+1)*n)*p);
      //Other terms
      lk=k*log(par)-s1;
      lk=exp(lk);
      t1=log(5.0)+le+k*lx-s1;
      t1=exp(t1)*exp(-(t-(3*k+1)*n)*p);
      t2=lk*(k+1)*p3/k;
      t2=t2*exp(-(t-(3*k+1)*n)*p);
      t3=lk*p*(2*n)*(k+1)*exp(p*((k+1)*(2*n)-t))/k;
      t4 = lk*abs(pow((t-k-4*n)/t,k)-1)*exp(-par)/k;
      //RESULT		  
      res=pbk+t1+t2+t3+t4; 
    }
  return res;
}

//===========================================================================================

int approx::psi_print_error_k(char* wf,int k,ofstream& fr)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m##########################\x1b[000m"<<endl
	  <<"\x1b[032m##### PSI ERROR IN K #####\x1b[000m"<<endl
	  <<"\x1b[032m##########################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"##########################"<<endl
    <<"##### PSI ERROR IN K #####"<<endl
    <<"##########################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p3,e1,le,b,res,x,lx,par;
  double s1;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
  	}
      else
	{
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1
	      e1=psiepsilon_one(V,n[i]);	      	     
	      //proposition psi
	      p3=proposition_psi(p,n[i],t,V);
	      if (k==0)
		{
		  res=p3;
		  if(VERBOSE)
		    {
		      cout<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
		    }
		  fr<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
		}
	      else
		{
		  	 
		  if(e1>=1)
		    {
		      if(VERBOSE)
			{
			  cout<<"Epsilon too large"<<endl<<endl;
			}
		      fr<<"Epsilon too large"<<endl<<endl;
		    }
		  else
		    {
		      le=log(e1);	      
		      //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		      x=par*(1+phi(n[i]));
		      lx=log(x);	
		      b=double(2)*t/double(n[i]);   
		      //Too large number of occurrences :
		      if (k>b)
			{
			 if(VERBOSE)
			    {
			      cout<<"Impossible : "<<endl
				  <<"a word of size "<<n[i]
				  << " and of period greater than "<< n[i]/2<<endl
				  <<"can not appear "<<k<<" times in a sequence of length "
				  <<t<<endl;
			      cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			    }
			  fr<<"Impossible : "<<endl
				  <<"a word of size "<<n[i]
				  << " and of period greater than "<< n[i]/2<<endl
				  <<"can not appear "<<k<<" times in a sequence of length "
				  <<t<<endl;
			  fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      else
			{
			  s1=0;
			  for (g=1;g<k;g++)
			    {
			      s1=s1+log(double(g));
			    }			  
			  res = psi_error_k(k,s1,lx,le,p3,n[i],p,par);
			  if(VERBOSE)
			    {
			      cout<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
			    }
			  fr<<"RESULT (error in "<<k<<") = "<<res<<endl<<endl;
			}
		    }
		}
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}

//===========================================================================================

double approx::psi_sum_errors_k
(int k,double s1,double lx, double le, double p3, int n, double p, double par)
{
  int j;
  double t=_length;
  double b= double(2)*t/double(n);
  double pr=pow(double(10),double(-300));
  double res=0;
  double result=pow(double(10),double(300));
  for (j=k;(j<=b) && (result>pr);j++)
    {
      result = psi_error_k(j,s1,lx,le,p3,n,p,par);
      res = res+result;
      s1=s1+log(double(j));
    }
  res = res + (b-j)*pr;
  return res;
}

//===========================================================================================

int approx::psi_print_sum_errors_k(char* wf,int k,ofstream& fr)
{

  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m#################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PSI SUM ERRORS OVER K #####\x1b[000m"<<endl
	  <<"\x1b[032m#################################\x1b[000m"<<endl<<endl;
    }
  fr<<endl;
  fr<<"#################################"<<endl
    <<"##### PSI SUM ERRORS OVER K #####"<<endl
    <<"#################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p3,e1,le,b,res,x,lx,par,s1;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words 
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
  	}
      else
	{
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	     if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1
	      e1=psiepsilon_one(V,n[i]);	      	     	      
	      //proposition psi
	      p3=proposition_psi(p,n[i],t,V);
	      if(e1>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		}
	      else
		{
		  le=log(e1);	      
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=par*(1+phi(n[i])); 		  
		  lx=log(x);	
		  b=double(2)*t/double(n[i]);   
		  //Too large number of occurrences :
		  if (k>b)
		    {
		      if(VERBOSE)
			{
			  cout<<"Impossible : "<<endl
			      <<"a word of size "<<n[i]
			      << " and of period greater than "<< n[i]/2<<endl
			      <<"can not appear "<<k<<" times in a sequence of length "
			      <<t<<endl;
			  cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      fr<<"Impossible : "<<endl
			<<"a word of size "<<n[i]
			<< " and of period greater than "<< n[i]/2<<endl
			<<"can not appear "<<k<<" times in a sequence of length "
			<<t<<endl;
		      fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
		    }
		  else
		    {
		      s1=0;
		      for (g=1;g<k;g++)
			{
			  s1=s1+log(double(g));
			}	
		      res = psi_sum_errors_k(k,s1,lx,le,p3,n[i],p,par);
		      double sm =sum_app(k,b,par);
		      if(VERBOSE)
			{
			  cout<<"RESULT (sum errors) = "<<res<<endl
			      <<"P("<<k<<" <= X <= "<<b<<"),"
			      <<" si X suit Poisson ("<<par<<") = "
			      <<sm<<endl<<endl;
			}
		      fr<<"RESULT (sum errors) = "<<res<<endl
			<<"P("<<k<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
			<<sm<<endl<<endl;
		    }
		}	       
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}

//===========================================================================================

int approx::psi_print_u(char* wf, double ds,ofstream& fr, ofstream& fu)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m#########################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PSI U (overrepresented words) #####\x1b[000m"<<endl
	  <<"\x1b[032m#########################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"#########################################"<<endl
    <<"##### PSI U (overrepresented words) #####"<<endl
    <<"#########################################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"####################################################"<<endl
    <<"##### u for psi mixing (overrepresented words) #####"<<endl
    <<"####################################################"<<endl<<endl;

  double t = _length;
 
  int k=0;
  int i,j,g,z;
  double p,p3,e1,le,b,res,result,x,s1,s2;
  double par,lx,pr,sm;
  pr=pow(double(10),double(-300));

  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {

      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      if(inBn2(V,n[i]))
  	{
 	 if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible"<<endl; 
  	}
      else
	{
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible"<<endl;
	    }  	  
	  else
	    {
	      //epsilon1
	      e1=psiepsilon_one(V,n[i]); 
	      //proposition psi
	      p3=proposition_psi(p,n[i],t,V);
	      //Error proposition 21	      
	      if(e1>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		  fu<<"impossible"<<endl;
		}
	      else
		{
		  le=log(e1);
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=par*(1+phi(n[i]));	     	     
		  lx=log(x);
		  b=double(2)*t/double(n[i]);     		  
		  res= pow(double(10),double(300));
		  //Search u from parameter (the mean) 
		  k=int(par)+1; 
		  s1=0;
		  for (g=1;g<k;g++)
		    {
		      s1=s1+log(double(g));
		    }
		  //TO RETURN U
		  result=pow(double(10),double(300));

 		  while (res>ds)
  		    {	
    		      j=k;
  		      res=sum_app(j,b,par);
  		      sm=res;
 		      s2=s1;
 		      if(sm<ds)
 			{
 			  while ((j<=b) && (result>pr) && (res<ds))
 			    {		      
 			      result=psi_error_k(j,s2,lx,le,p3,n[i],p,par);
 			      res=res+result; 
 			      s2=s2+log(double(j)); 
 			      j++;
 			    }
 			  if (result<pr)
 			    {
 			      res=res+(b-j+1)*pr;			 
 			    }
 			}
  		      s1=s1+log(double(k));  		      
  		      k++;		       
  		    }
		  if(VERBOSE)
		    {
		      cout<<"RESULT (sum errors) = "<<res<<endl
			  <<"P("<<k-1<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
			  <<sm<<endl<<endl;
		    }
		  fr<<"RESULT (sum errors) = "<<res<<endl
		    <<"P("<<k-1<<" <= X <= "<<b<<"), if X follows Poisson ("<<par<<") = "
		    <<sm<<endl<<endl;
		  fu<<k-1<<endl;
		}
	    }  			       
	}
      delete[] V;    
    }
  delete[] n; 
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                  PSI MIXING                 @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@     underrepresented words     @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::psi_k_sum_errors
(int k,double lx, double le, double p3, int n, double p, double par)
{ 
  int j;
  double res=0;
  double result;
  double s1=0;
  result = psi_error_k(0,s1,lx,le,p3,n,p,par);
  res = res+result;
  for (j=1;j<k+1;j++)
    {
      result = psi_error_k(j,s1,lx,le,p3,n,p,par);
      res = res+result;
      s1=s1+log(double(j));
    }
  return res;
}

//===========================================================================================

int approx::psi_print_k_sum_errors(char* wf,int k,ofstream& fr)
{if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m##################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PSI SUM ERRORS UNDER K #####\x1b[000m"<<endl
	  <<"\x1b[032m##################################\x1b[000m"<<endl<<endl;
    }
  fr<<endl;
  fr<<"##################################"<<endl
    <<"##### PSI SUM ERRORS UNDER K #####"<<endl
    <<"##################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;
  double t=_length;
 
  int i,j,g;
  double p,p3,e1,le,b,res,x,lx,par;
   
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words  
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();

  //################## Loop on i : study of (i+1)th word ##################//
       
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	} 
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      if(inBn2(V,n[i]))
  	{ 
	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
  	}
      else
	{
	  p= probawll(n[i],V,n[i]);  
	  par=t*p;
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	    }  
	  else
	    {  
	      //epsilon1
	      e1=psiepsilon_one(V,n[i]);	      	     	      
	      //proposition psi
	      p3=proposition_psi(p,n[i],t,V);
	      if(e1>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		}
	      else
		{
		  le=log(e1);	      
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=par*(1+phi(n[i])); 		  
		  lx=log(x);	
		  b=double(2)*t/double(n[i]);   
		  //Too large number of occurrences :
		  if (k>b)
		    {
		      if(VERBOSE)
			{
			  cout<<"Impossible : "<<endl
			      <<"a word of size "<<n[i]
			      << " and of period greater than "<< n[i]/2<<endl
			      <<"can not appear "<<k<<" times in a sequence of length "
			      <<t<<endl;
			  cout<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
			}
		      fr<<"Impossible : "<<endl
			<<"a word of size "<<n[i]
			<< " and of period greater than "<< n[i]/2<<endl
			<<"can not appear "<<k<<" times in a sequence of length "
			<<t<<endl;
		      fr<<"No error, P(N = "<<k<<") = 0."<<endl<<endl;
		    }
		  else
		    {
		     
		      res = psi_k_sum_errors(k,lx,le,p3,n[i],p,par);
		      double sm =sum_app(0,k,par);
		      if(VERBOSE)
			{
			  cout<<"RESULT (sum error) = "<<res<<endl
			      <<"P(0 <= X <= "<<k<<"),"
			      <<" if X follows Poisson ("<<par<<") = "
			      <<sm<<endl<<endl;
			}
		      fr<<"RESULT (sum errors) = "<<res<<endl
			<<"P(0 <= X <= "<<k<<"), if X follows Poisson ("<<par<<") = "
			<<sm<<endl<<endl;
		    }
		}	       
	    }
	}
      delete[] V;
    }
  delete[] n;
  return 1;
}

//===========================================================================================

int approx::psi_u_print(char* wf, double ds,ofstream& fr, ofstream& fu)
{if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[032m##########################################\x1b[000m"<<endl
	  <<"\x1b[032m##### PSI U (underrepresented words) #####\x1b[000m"<<endl
	  <<"\x1b[032m##########################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"##########################################"<<endl
    <<"##### PSI U (underrepresented words) #####"<<endl
    <<"##########################################"<<endl<<endl; 
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"#####################################################"<<endl
    <<"##### u for psi mixing (underrepresented words) #####"<<endl
    <<"#####################################################"<<endl<<endl;

  double t = _length;
 
  int k=0;
  int i,j,g,z;
  double p,p3,e1,le,res,result,x,s1;
  double par,lx,sm;
  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words  
  mo.getline(T,256);   
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {

      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{ 
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<" : "<<V<<"\x1b[000m"<<endl;
	}
      fr<<"WORD "<<i+1<<" : "<<V<<endl;
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      if(inBn2(V,n[i]))
  	{
 	  if(VERBOSE)
 	    {
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible"<<endl;
  	}
      else
	{
	  if(VERBOSE)
	    {
	      cout<<"Probability of the word = "<<p<<endl; 
	      cout<<"Parameter of the Poisson law = "<<par<<endl; 
	    }
	  fr<<"Probability of the word = "<<p<<endl; 
	  fr<<"Parameter of the Poisson law = "<<par<<endl;
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible"<<endl;
	    }  	  
	  else
	    {
	      //epsilon1
	      e1=psiepsilon_one(V,n[i]); 
	      //proposition psi
	      p3=proposition_psi(p,n[i],t,V);
	      //Error proposition 21	      
	      if(e1>=1)
		{
		  if(VERBOSE)
		    {
		      cout<<"Epsilon too large"<<endl<<endl;
		    }
		  fr<<"Epsilon too large"<<endl<<endl;
		  fu<<"impossible"<<endl;
		}
	      else
		{
		  le=log(e1);
		  //x power k = lambda_k = max(1,((p+phi(la))*t)^k)
		  x=par*(1+phi(n[i]));	     	     
		  lx=log(x);
		  res= 0;
		  //Search u from 0
		  k=0; 
		  s1=0;
		  result=psi_error_k(0,s1,lx,le,p3,n[i],p,par);
		  res=res+result+sum_app(0,0,par);
		  k++;
		  //TO RETURN U
		  while (res<ds)
  		    {	
    		      res=sum_app(0,k,par);
 		      sm=res; 		      
 		      if(sm<ds)
 			{
			  j=1;
 			  while ((j<=k) && (res<ds))
 			    {		      
 			      result=psi_error_k(j,s1,lx,le,p3,n[i],p,par);
 			      res=res+result; 
 			      s1=s1+log(double(j)); 
 			      j++;
 			    }
			}
  		      k++;		       
  		    }
		  if(VERBOSE)
		    {
		      cout<<"RESULT (sum errors) = "<<res<<endl
			  <<"P(0 <= X <= "<<k-1<<"), if X follows Poisson ("<<par<<") = "
			  <<sm<<endl<<endl;
		    }
		  fr<<"RESULTAT (sum errors) = "<<res<<endl
		    <<"P(0 <= X <= "<<k-1<<"), if X follows Poisson ("<<par<<") = "
		    <<sm<<endl<<endl;
		  fu<<k-1<<endl;
		}
	    }  			       
	}
      delete[] V;    
    }
  delete[] n; 
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                 CHEN-STEIN                  @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@     overrepresented words      @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::deltacs(double p, int n)
{
  double C= 1-phi(double(3*n));
  double lnu=log(_nu);
  double x= -(5+2*phi(n))*p*C;
  double y=4*_K*lnu;
  double z=log(x/y)/lnu;
  return z;
}

//===========================================================================================

double approx::eps2cs(double p, int n)
{
  double d=deltacs(p,n);
  double eps= (5+2*phi(n))*d*p+4*phi(d)/(1-phi(3*n));
  return eps;
}

//===========================================================================================

double approx::chenstein_over_ma(char* wf, double ds,ofstream& fr, ofstream& fu)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[001;034m#####################################\x1b[000m"<<endl
	  <<"\x1b[001;034m##### Chen-Stein (Miguel Abadi) #####\x1b[000m"<<endl
	  <<"\x1b[001;034m#####################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"#####################################"<<endl
    <<"##### Chen-Stein (Miguel Abadi) #####"<<endl
    <<"#####################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"##########################################################"<<endl
    <<"##### u for Chen-Stein Abadi (overrepresented words) #####"<<endl
    <<"##########################################################"<<endl<<endl;

  double t = _length;

  int k=0;

  int i,j,l;
  double p,eps1,eps2,CS;
  double par,pr,sm,seuil; 
  pr=pow(double(10),double(-300));

  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words   
  mo.getline(T,256); 
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<", "<<V<<" : \x1b[000m";
	}
      fr<<"WORD "<<i+1<<", "<<V<<" : "; 
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      
      if(inBn2(V,n[i]))
       	{
	  if(VERBOSE)
 	    {
	      cout<<endl;
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
	  fr<<endl;
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible";
       	} 
      else
	{
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<endl;
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<endl;
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<endl;
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<endl;
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible";
	    }  	  
	  else
	    {
	      //epsilon
	      eps1=psiepsilon_one(V,n[i]); 	       
	      eps2=eps2cs(p,n[i]);
	      CS=par*(eps1+eps2);
	      if(VERBOSE)
		{ 
		  cout<<CS<<endl;	     	     	
		}
	      fr<<CS<<endl;	     
	      //Return u such that sum_app + CS<ds
	      if (CS>ds)
		{
		  fu<<"impossible"; 
		}
	      else
		{
		  l=0;
		  sm=0;
		  seuil=ds-CS;
		  while (1-sm>seuil)
		    {
		      sm=sm+poiss(l,par);
		      l=l+1;
		    }
		  fu<<l;  
		}   	 
	    }
	}


      fu<<"  [Chenstein = "<<CS<<"]"<<endl;


      delete[] V;    
    }
  delete[] n; 
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@    Count of Words   @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

int approx::count_words(char* wf, char* sf, char* fa, ofstream& fu)
{
  if(VERBOSE)
    {
      cout<<endl
	  <<"\x1b[001;034m#####################################\x1b[000m"<<endl
	  <<"\x1b[001;034m##   COUNT OF WORDS IN SEQUENCES   ##\x1b[000m"<<endl
	  <<"\x1b[001;034m#####################################\x1b[000m"<<endl
	  <<endl;
      cout<<"see u_file"<<endl;
    }
  fu<<endl
    <<"#####################################"<<endl
    <<"##   COUNT OF WORDS IN SEQUENCES   ##"<<endl
    <<"#####################################"<<endl
    <<endl;
  SequenceSet seqs(sf,fa,_order);
  ifstream seq(sf);
  int nbw,nb,i,j,c;
  c=1;
  char * S, * W;
  S=new char[256]; 
  W=new char[256];
  ifstream words(wf); 
  words>>nbw;
  words.close();
  int * NB= new int[nbw];
  for (i=0;i<nbw;i++)
    {
      NB[i]=0;
    }
  while (!seq.eof())
    { 
      Sequence SEQ(seqs.get_sequence(c-1));
      words.open(wf);      
      words>>nbw;
      seq>>S;
      fu<<endl<<"Sequence "<<c<<" : "<<S<<endl; 
      for (i=0;i<nbw;i++)
	{
 	  words>>W;
 	  string w(W);
 	  nb=0;
 	  const string s = SEQ.strseq();
	  j=0;
  	  while(j<seqs.tell_length_seq(c-1) && j>=0)
  	    {
 	      j=s.find(w,j+1);
	      nb++;
 	    }
	  nb--;
	  NB[i]+=nb;
 	  fu<<"WORD "<<W<<" : "<<nb<<endl; 
	}
      words.close();
      c++;
    }
  words.open(wf);
  words>>nbw;
  fu<<endl<<"All sequences : "<<endl; 
  for (i=0;i<nbw;i++)
    {
      words>>W;
      fu<<"WORD "<<W<<" : "<<NB[i]<<endl; 
    }
  words.close();
  delete[] NB;
  delete[] S;
  delete[] W;
  return 1;
}

//===========================================================================================

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                 CHEN-STEIN                  @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@                                             @@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@    Underrepresented words      @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@                                @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

//===========================================================================================

double approx::chenstein_under_ma(char* wf, double ds,ofstream& fr, ofstream& fu)
{
  if(VERBOSE)
    {
      cout<<endl;
      cout<<"\x1b[001;034m#####################################\x1b[000m"<<endl
	  <<"\x1b[001;034m##### Chen-Stein (Miguel Abadi) #####\x1b[000m"<<endl
	  <<"\x1b[001;034m#####################################\x1b[000m"<<endl<<endl; 
    }
  fr<<endl;
  fr<<"#####################################"<<endl
    <<"##### Chen-Stein (Miguel Abadi) #####"<<endl
    <<"#####################################"<<endl<<endl;
  if(VERBOSE)
    {
      cout<<"\x1b[001mPsi mixing\x1b[000m"<<endl<<endl; 
    }
  fr<<"Psi mixing"<<endl<<endl;

  fu<<endl;
  fu<<"###########################################################"<<endl
    <<"##### u for Chen-Stein Abadi (underrepresented words) #####"<<endl
    <<"###########################################################"<<endl<<endl;

  double t = _length;

  int k=0;

  int i,j,l;
  double p,eps1,eps2,CS;
  double par,pr,sm,seuil; 
  pr=pow(double(10),double(-300));

  int nb;
  ifstream mo;
  mo.open(wf);  
  mo>>nb;
  int* n=new int[nb]; 
  char T[256];
  //Length of words  
  mo.getline(T,256); 
  for(i=0; i<nb; i++)
    {
      mo.getline(T,256);
      n[i]=strlen(T);
    }
  mo.close();
  //################## Loop on i : study of (i+1)th word ##################//
  
  for(i=0;i<nb;i++)
    {
      //word i+1,
      char*V=new char[n[i]+1];
      for (j=0;j<n[i]+1;j++)
	{
	  char* A=new char[n[i]+1];
	  give_word(A,wf,i);
	  V[j]=A[j];
	  delete[] A;
	}
      p= probawll(n[i],V,n[i]);
      par=t*p;
      if(VERBOSE)
	{
	  cout<<"\x1b[036mWORD "<<i+1<<", "<<V<<" : \x1b[000m";
	}
      fr<<"WORD "<<i+1<<", "<<V<<" : "; 
      fu<<"WORD "<<i+1<<" ("<<V<<") : u = ";
      
      if(inBn2(V,n[i]))
       	{
	  if(VERBOSE)
 	    {
	      cout<<endl;
 	      cout<<"This word do not satisfy hypothesis :"<<endl
 		  <<"period greater than half of its length."<<endl<<endl; 
 	    }
	  fr<<endl;
  	  fr<<"This word do not satisfy hypothesis :"<<endl
	    <<"period greater than half of its length."<<endl<<endl; 
	  fu<<"impossible";
       	} 
      else
	{
	  //p must be greater than 0
	  //tp must be smaller than 500, else (tp)^k/k! too large
	  if ((p==0) || (par > 500))
	    {
	      if(p==0)
		{
		  if(VERBOSE)
		    {
		      cout<<endl;
		      cout<<"p=0. Transition matrix contains 0"<<endl<<endl;
		    }
		  fr<<endl;
		  fr<<"p=0. Transition matrix contains 0"<<endl<<endl;
		}
	      else
		{
		  if(VERBOSE)
		    {
		      cout<<endl;
		      cout<<"Too large parameter"<<endl<<endl;
		    }
		  fr<<endl;
		  fr<<"Too large parameter"<<endl<<endl;
		}
	      fu<<"impossible";
	    }  	  
	  else
	    {
	      //epsilon
	      eps1=psiepsilon_one(V,n[i]); 	       
	      eps2=eps2cs(p,n[i]);
	      CS=par*(eps1+eps2);
	      if(VERBOSE)
		{ 
		  cout<<CS<<endl;	     	     	
		}
	      fr<<CS<<endl;	     
	      //Return u such that sum_app + CS<ds
	      if (CS>ds)
		{
		  fu<<"impossible"; 
		}
	      else
		{
		  l=0;
		  sm=0;
		  seuil=ds-CS;
		  while (sm<seuil)
		    {
		      sm=sm+poiss(l,par);
		      l=l+1;
		    }
		  fu<<l-2;  
		}   	 
	    }
	}


      fu<<"  [Chenstein = "<<CS<<"]"<<endl;


      delete[] V;    
    }
  delete[] n; 
  return 1;
}


